/***** spin: vars.c *****/

/* Copyright (c) 1991-1998 by Lucent Technologies - Bell Laboratories     */
/* All Rights Reserved.  This software is for educational purposes only.  */
/* Permission is given to distribute this code provided that this intro-  */
/* ductory message is not removed and no monies are exchanged.            */
/* No guarantee is expressed or implied by the distribution of this code. */
/* Software written by Gerard J. Holzmann as part of the book:            */
/* `Design and Validation of Computer Protocols,' ISBN 0-13-539925-4,     */
/* Prentice Hall, Englewood Cliffs, NJ, 07632.                            */
/* Send bug-reports and/or questions to: gerard@research.bell-labs.com    */

/* dSpin extension written by Radu Iosif. Feedback: iosif@athena.polito.it */


#include "spin.h"
#ifdef PC
#include "y_tab.h"
#else
#include "y.tab.h"
#endif

extern Ordered	*all_names;
extern RunList	*X, *LastX;
extern Symbol	*Fname;
extern char	Buf[];
extern int	lineno, depth, verbose, xspin, limited_vis;
extern int	analyze, jumpsteps, nproc, columns;
extern short	no_arrays, Have_claim, Skip_claim;
extern void	sr_mesg(FILE *, int, int);
extern void	sr_buf(int, int);

extern Symbol  *struct_sym(Lextok *, Symbol *);  
extern void     xputname(FILE *, char *, Lextok *, int, char *);
extern void     Sclear();

static int	getglobal(Lextok *);
static int      getglobal_lval(Lextok *);
static int	setglobal(Lextok *, int);
static int      setglobal_lval(Lextok *, int, int);
static int      setglobal_func(Lextok *, Sequence *);
static int	maxcolnr = 1;



int
get_val(Symbol *s, int n)
{
  if (!s->val)
    fatal("use of null pointer '%s'", s->name);

  if (n < 0 || n >= s->val[0]) 
    {    
      printf("spin: index out of range '%s[%d]' - actual size is %d\n", 
	     s->name, n, s->val[0]);
      fatal("indexing array '%s'", s->name);
    }

  if (s->type & POINTER && s->nel > 1)
    {
      if (!(int *)s->val[n + 1]) {
	char buf[256];
	sprintf(buf, "use of null pointer '%s[%d]'", s->name, n);
	fatal(buf, 0);
      }

      return ((int *)s->val[n + 1]) [1];
    }

  return s->val[n + 1];
}


int
get_lval(Symbol *s, Lextok *ix)
{
  if (ix)
    {
      int n = eval(ix);
      if (n < 0 || n >= s->val[0])
	{
	  printf("spin: index out of range '%s[%d]' - actual size is %d\n", 
		 s->name, n, s->val[0]);
	  fatal("indexing array '%s'", s->name);	  
	}

      return s->val[n + 1];
    }
  else
    return (int)s->val;
}


Lextok *
get_Sval(Symbol *s, int n)
{
  if (!s->Sval)
    fatal("use of null pointer '%s'", s->name);

  if (n < 0 || n >= (int)s->Sval[0])
    {
      printf("spin: index out of range '%s[%d]' - actual size is %d\n", 
	     s->name, n, s->val[0]);
      fatal("indexing array '%s'", s->name);
    }

  return s->Sval[n + 1];
}


void
set_val(Symbol *s, int n, int m)
{
  if (!s->val)
    fatal("assignment to null pointer '%s'", s->name);

  if (n < 0 || n >= s->val[0]) 
    {    
      printf("spin: index out of range '%s[%d]' - actual size is %d\n", 
	     s->name, n, s->val[0]);
      fatal("indexing array '%s'", s->name);
    }
  
  if (s->type & POINTER && s->nel > 1)
    {
      if (!(int *)s->val[n + 1]) {
	char buf[256];
	sprintf(buf, "assignment to null pointer '%s[%d]'", s->name, n);
        fatal(buf, 0);
      }

      ((int *)s->val[n + 1]) [1] = m;
    }
  else
    s->val[n + 1] = m;
}


void
set_lval(Symbol *s, Lextok *ix, int m)
{
  if (ix)
    {
      int n = eval(ix);

      if (n < 0 || n >= s->val[0]) 
	{    
	  printf("spin: index out of range '%s[%d]' - actual size is %d\n", 
		 s->name, n, s->val[0]);
	  fatal("indexing array '%s'", s->name);
	}

      s->val[n + 1] = m;
    }
  else
    s->val = (int *)m;
}


void
set_Sval(Symbol *s, int n, Lextok *m)
{
  if (!s->Sval)
    fatal("assignment to null pointer '%s'", s->name);

  if (n < 0 || n >= (int)s->Sval[0])
    {
      printf("spin: index out of range '%s[%d]' - actual size is %d\n", 
	     s->name, n, s->val[0]);
      fatal("indexing array '%s'", s->name);
    }

  s->Sval[n + 1] = m;
}


int
getval(Lextok *sn)
{	
  Symbol *s = sn->sym;
  
  if (strcmp(s->name, "_") == 0)
    fatal("attempt to read value of '_'", 0);
  
  if (strcmp(s->name, "_last") == 0)
    return (LastX) ? LastX->pid : 0;
  
  if (strcmp(s->name, "_p") == 0)
    return (X && X->pc) ? X->pc->seqno : 0;
  
  if (strcmp(s->name, "_pid") == 0)
    {	
      if (!X) return 0;
      return X->pid - Have_claim;		
    }
  
  if (findloc(s))
    return getlocal(sn);
  
  return getglobal(sn);
}

int
getlval(Lextok *sn)
{
  Symbol *s = sn->sym;

  if (findloc(s))
    return getlocal_lval(sn);

  return getglobal_lval(sn);
}

int
setval(Lextok *v, int n)
{
  if (findloc(v->sym))
    return setlocal(v, n);

  return setglobal(v, n);
}

int 
setlval(Lextok *v, int n, int heap)
{
  if (findloc(v->sym))
    return setlocal_lval(v, n, heap);

  return setglobal_lval(v, n, heap);
}

int
setfunc(Lextok *v, Sequence *body)
{
  if (findloc(v->sym))
    return setlocal_func(v, body);
  
  return setglobal_func(v,  body);
}

void
rm_selfrefs(Symbol *s, Lextok *i)
{
	if (!i) return;

	if (i->ntyp == NAME
	&&  strcmp(i->sym->name, s->name) == 0
	&& (	(!i->sym->context && !s->context)
	   ||	( i->sym->context &&  s->context
		&& strcmp(i->sym->context->name, s->context->name) == 0)))
	{	lineno  = i->ln;
		Fname   = i->fn;
		non_fatal("self-reference initializing '%s'", s->name);
		i->ntyp = CONST;
		i->val  = 0;
	} else
	{	rm_selfrefs(s, i->lft);
		rm_selfrefs(s, i->rgt);
	}
}

int
checkvar(Symbol *s, int n)
{	
  int i, oln = lineno;	/* calls on eval() change it */
  Symbol *ofnm = Fname;

  if (!(s->type & POINTER && s->nel == 1)
      && (n >= s->nel || n < 0))
    {	
      printf("spin: indexing %s[%d] - size is %d\n", s->name, n, s->nel);
      non_fatal("indexing array \'%s\'", s->name);      return 0;
    }

  if (s->type == 0)
    fatal("undeclared variable %s", s->name);
}

int
ini_var(Symbol *s)
{
  int i, size;

  size = var_size(s);

  if (s->type == FTYPE)
    {
      if (!s->Sval) {
	s->Sval = (Lextok **) emalloc(size + sizeof(Lextok *));
	s->Sval[0] = (Lextok *) s->nel;
      }
    }
  else
    if (s->type & POINTER && s->nel > 1)
      {
	if (s->type & STRUCT)
	  {
	    if (!s->Sval) {
	      s->Sval = (Lextok **) emalloc(size + sizeof(Lextok *));
	      s->Sval[0] = (Lextok *) s->nel;
	    }
	  }
	else
	  {
	    if (!s->val) {
	      s->val = (int *) emalloc(size + sizeof(int));
	      s->val[0] = s->nel;
	    }
	  }
      }
    else
      if (!(s->type & POINTER) && !s->val) {
	s->val = (int *) emalloc(size + sizeof(int));
	s->val[0] = s->nel;
      }

  for (i = 0; i < s->nel; i ++)
    if (s->type != CHAN)
      {	
	if (s->ini)
	  {
	    rm_selfrefs(s, s->ini);
	    if (s->type == FTYPE)
	      set_Sval(s, i, (Lextok *)eval(s->ini));
	    else
	      {
		if (s->type & POINTER)
		  {
		    extern void comment(FILE *, Lextok *, int);

		    printf("line %d:\n", s->ini->ln);
		    printf("=====> instead of \n");
		    printf("\t<type> &%s = ", s->name);
		    comment(stdout, s->ini, 0);
		    printf(";\n");
		    printf("=====> always use  \n");
		    printf("\t<type> &%s;\n", s->name);
		    printf("\t%s = ", s->name);
		    comment(stdout, s->ini, 0);
		    printf(";\n");

		    alldone(1);
		  }
		else
		  set_val(s, i, eval(s->ini));
	      }
	  }
      } 
    else 
      s->val[i + 1] = qmake(s);
}

int
is_on_heap(Lextok *n)
{
  Symbol *s;

  if (!n)
    return 0;

  if (n->ntyp == NEW)
    return 1;

  s = struct_sym(n, n->sym);
  return s ? (s->hidden & 64) : 0;
}


void initglobals()
{
  Ordered *walk;
  Symbol *sp;
  Lextok *n;
  Value v;
  int type;
  
  n = (Lextok *) emalloc(sizeof(Lextok));
  n->ntyp = NAME;

  for (walk = all_names; walk; walk = walk->next)
    {
      sp = walk->entry;

      if (sp 
	  && !sp->context 
	  && !sp->owner
	  && !(sp->hidden & 1)
	  && !(sp->hidden & 128)
	  && sp->Nid >= 0
	  && (sp->type == UNSIGNED
	      ||  sp->type == PREDEF
	      ||  sp->type == BIT
	      ||  sp->type == MTYPE
	      ||  sp->type == BYTE
	      ||  sp->type == CHAN
	      ||  sp->type == SHORT
	      ||  sp->type == INT
	      ||  sp->type == STRUCT
	      ||  sp->type == FTYPE
	      ||  sp->type & POINTER))
	{
	  /* Allocate run-time space */
	  if (sp->type == STRUCT)
	    ini_struct(sp);
	  else
	    ini_var(sp);
	}
    }
}


static int 
getglobal(Lextok *sn)
{	
  Symbol *s = sn->sym;
  int i, n;

  n = eval(sn->lft);

  if (s->type == 0 && X && (i = find_lab(s, X->n, 0)))
    {	
      printf("findlab through getglobal on %s\n", s->name);
      return i;	/* can this happen? */
    }

  if (s->type & STRUCT)
    return Rval_struct(sn, s);

  if (checkvar(s, n))
    {
      if (s->type == FTYPE)       
	return (int)get_Sval(s, n);
      else
	return cast_val(s->type, get_val(s, n), s->nbits);
    }

  return 0;
}


static int getglobal_lval(Lextok *n)
{
  Symbol *s = n->sym;

  if (s->type & STRUCT)
    return Rval_lstruct(n, s);
  else
    return get_lval(s, n->lft);
}


int var_size(Symbol *s)
{
  Lextok *n, *t;
  int sum = 0;

  if (s->type & POINTER)
    return s->nel * sizeof(void *);

  switch (s->type)
    {
    case BIT:
    case UNSIGNED:
      return s->nel;

    case BYTE:
    case SHORT:
    case INT:
      return s->nel * (s->type & ~POINTER) / 8;

    case MTYPE:
    case CHAN:
    case PREDEF:
      return s->nel * sizeof(int);

    case STRUCT:
    case FTYPE:
      return s->nel * sizeof(Lextok *);

    default:
      fatal("cannot happen (var_size(%s))", (s && s->name) ? s->name : NULL);
    }
}


int cast_val(int t, int v, int w)
{	
  int i=0; short s=0; unsigned char u=0;
  
  if (t & INT || t == CHAN) 
    i = v;

  else if (t & SHORT) s = (short) v;
  else if (t & BYTE || t & MTYPE)  u = (unsigned char)v;
  else if (t & BIT)   u = (unsigned char)(v&1);
  else if (t & UNSIGNED) {	
    if (w == 0)
      fatal("cannot happen, cast_val", (char *)0);

    u = (unsigned char)(v& ((1<<w)-1));
  }

  if (v != i+s+u)
    {	
      char buf[32]; 
      
      sprintf(buf, "%d->%d (%d)", v, i+s+u, t);
      non_fatal("value (%s) truncated in assignment", buf);
    }

  return (int)(i+s+u);
}


static int
setglobal(Lextok *v, int m)
{
  int n;

  if (v->sym->type & STRUCT)
    (void) Lval_struct(v, v->sym, m);
  else 
    {
      n = eval(v->lft);
      if (checkvar(v->sym, n))
	{	
	  if (v->sym->nbits > 0)
	    m = (m & ((1<<v->sym->nbits)-1));	

	  set_val(v->sym, n, m);
	  v->sym->setat = depth;
	}	
    }
  
  return 1;
}

static int
setglobal_lval(Lextok *v, int m, int heap)
{
  if (v->sym->type & STRUCT)
    Lval_lstruct(v, v->sym, m, heap);
  else 
    {
      set_lval(v->sym, v->lft, m);

      if (heap)
	v->sym->hidden &= 64;
    }

  return 1;
}


static int
setglobal_func(Lextok *v, Sequence *body)
{
  int n = eval(v->lft);
  if (!(v->sym->type & POINTER && v->sym->nel == 1)
      && (n >= v->sym->nel || n < 0))
    {
      printf("spin: indexing %s[%d] - size is %d\n", 
	     v->sym->name, n, v->sym->nel);
      non_fatal("indexing array \'%s\'", v->sym->name);      
    } 
  else {
    if (v->sym->type & STRUCT)
      (void) Lval_fstruct(v, v->sym, body);
    else
      if (checkvar(v->sym, n))
	set_Sval(v->sym, n, (Lextok *)body);
  }
  
  return 1;
}


void dumpclaims(FILE *fd, int pid, char *s)
{	
  extern Lextok *Xu_List; extern int Pid;
  extern short terse;
  Lextok *m; int cnt = 0; int oPid = Pid;

  for (m = Xu_List; m; m = m->rgt)
    if (strcmp(m->sym->name, s) == 0)
      {	
	cnt = 1;
	break;
      }

  if (cnt == 0) 
    return;

  Pid = pid;
  fprintf(fd, "#ifndef XUSAFE\n");

  for (m = Xu_List; m; m = m->rgt)
    {	
      if (strcmp(m->sym->name, s) != 0)
	continue;

      no_arrays = 1;
      putname(fd, "  setq_claim(", m->lft, 0, "", 1);
      no_arrays = 0;
      fprintf(fd, ", %d, ", m->val);
      terse = 1;
      xputname(fd, "\"", m->lft, 0, "\", h, ");
      terse = 0;
      fprintf(fd, "\"%s\");\n", s);
    }

  fprintf(fd, "#endif\n");
  Pid = oPid;
}


void 
dumpglobals(void)
{	
  Ordered *walk;
  Lextok *dummy;
  Symbol *sp;
  int j;

  for (walk = all_names; walk; walk = walk->next)
    {	
      sp = walk->entry;
      
      if (!sp->type 
	  || sp->hidden & 1
	  || sp->hidden & 128
	  || sp->context 
	  || sp->owner
	  || sp->type == PROCTYPE 
	  || sp->type == PREDEF
	  || (sp->type == MTYPE && ismtype(sp->name)))
	continue;
	
      if (limited_vis && !(sp->hidden & 2))
	continue;

      if (sp->type & STRUCT && sp->Sval)
	{
	  Sclear();
	  dump_struct(sp, sp->name, 0);
	  continue;
	}
	
      for (j = 0; j < sp->nel; j++)
	{	
	  int prefetch;
		
	  if (sp->type == CHAN)
	    {	
	      doq(sp, j, 0);
	      continue;
	    }
		
	  if ((verbose&4)
	      && (sp->setat < depth
		  && jumpsteps != depth))
	    continue;
		
	  dummy = nn(ZN, NAME, nn(ZN,CONST,ZN,ZN), ZN);
	  dummy->sym = sp;
	  dummy->lft->val = j;
	  
	  /* in case of cast_val warnings, do this first: */
	  
	  if (dummy->sym->val)
	    prefetch = getglobal(dummy);
	  else
	    prefetch = 0;

	  if (!limited_vis)
	    {	
	      printf("\t\t%s", sp->name);

	      if (sp->nel > 1) 
		printf("[%d]", j);

	      printf(" = ");
	      sr_mesg(stdout, prefetch, sp->type == MTYPE);
	      printf("\n");
	    } 
	  else
	    {	
	      int colpos;
	      
	      Buf[0] = '\0';
	      if (!xspin)
		{	
		  if (columns == 2)
		    sprintf(Buf, "~G%s = ", sp->name);
		  else
		    sprintf(Buf, "%s = ", sp->name);
		}
	      
	      sr_buf(prefetch, sp->type == MTYPE);

	      if (sp->colnr == 0)
		{	
		  sp->colnr = maxcolnr;
		  maxcolnr = 1 + (maxcolnr % 10);
		}

	      colpos = nproc+sp->colnr-1;
	      
	      if (columns == 2)
		{	
		  pstext(colpos, Buf);
		  continue;
		}
		
	      if (!xspin)
		{	
		  printf("\t\t%s\n", Buf);
		  continue;
		}
		
	      printf("MSC: ~G %s %s\n", sp->name, Buf);
	      printf("%3d:\tproc %3d (TRACK) line   1 \"var\" ", 
		     depth, colpos);

	      printf("(state 0)\t[printf('MSC: globvar\\\\n')]\n");
	      printf("\t\t%s", sp->name);
	      
	      if (sp->nel > 1) 
		printf("[%d]", j);

	      printf(" = %s\n", Buf);
	    }	
	}	
    }
}


void
dumplocal(RunList *r)
{	
  Lextok *dummy;
  Symbol *z, *s = r->rt_seq->this->symtab;
  int i;

  for (z = s; z; z = z->next)
    {	
      if (z->hidden & 1
	  || s->type == LABEL
	  || (limited_vis && !(z->hidden&2)))
	continue;
	
      if (z->type & STRUCT && z->Sval)
	{	
	  Sclear();
	  dump_struct(z, z->name, r);
	  continue;
	}
	
      for (i = 0; i < z->nel; i++)
	{	
	  if (z->type == CHAN)
	    {	
	      doq(z, i, r);
	      continue;
	    }
	  
	  if ((verbose&4)
	      &&  (z->setat < depth
		   &&   jumpsteps != depth))
	    continue;

	  dummy = nn(ZN, NAME, nn(ZN,CONST,ZN,ZN), ZN);
	  dummy->sym = z;
	  dummy->lft->val = i;
	  
	  if (!limited_vis)
	    {	
	      printf("\t\t%s(%d):%s", r->n->name, r->pid, z->name);

	      if (z->nel > 1) printf("[%d]", i);
	      printf(" = ");
	      sr_mesg(stdout, getval(dummy), z->type == MTYPE);
	      printf("\n");
	    } 
	  else
	    {	
	      int colpos;
		
	      Buf[0] = '\0';
	      if (!xspin)
		{	
		  if (columns == 2)
		    sprintf(Buf, "~G%s(%d):%s = ",
			    r->n->name, r->pid, z->name);
		  else
		    sprintf(Buf, "%s(%d):%s = ",
			    r->n->name, r->pid, z->name);
		}
		
	      sr_buf(getval(dummy), z->type==MTYPE);
	      
	      if (z->colnr == 0)
		{	
		  z->colnr = maxcolnr;
		  maxcolnr = 1 + (maxcolnr % 10);
		}

	      colpos = nproc+z->colnr-1;
	      
	      if (columns == 2)
		{	
		  pstext(colpos, Buf);
		  continue;
		}
		
	      if (!xspin)
		{	
		  printf("\t\t%s\n", Buf);
		  continue;
		}
		
	      printf("MSC: ~G %s(%d):%s %s\n",
		     r->n->name, r->pid, z->name, Buf);
	      
	      printf("%3d:\tproc %3d (TRACK) line   1 \"var\" ",depth, colpos);
	      printf("(state 0)\t[printf('MSC: locvar\\\\n')]\n");
	      printf("\t\t%s(%d):%s", r->n->name, r->pid, z->name);
	      
	      if (z->nel > 1) 
		printf("[%d]", i);

	      printf(" = %s\n", Buf);
	    }	
	}	
    }
}
